import re
from decaychain.Nuclide import Nuclide

class DataRow:
    def __init__(self):
        self.COLUMN_NAMES = [
            'nuclide_name',
            'halflife',
            'decay_mode',
            'pointer_1',
            'pointer_2',
            'pointer_3',
            'pointer_4',
            'daughter_nuclide_1_name',
            'daughter_nuclide_1_record_i',
            'daughter_nuclide_1_branch_i',
            'daughter_nuclide_2_name',
            'daughter_nuclide_2_record_i',
            'daughter_nuclide_2_branch_i',
            'daughter_nuclide_3_name',
            'daughter_nuclide_3_record_i',
            'daughter_nuclide_3_branch_i',
            'daughter_nuclide_4_name',
            'daughter_nuclide_4_record_i',
            'daughter_nuclide_4_branch_i',
            'energy_alpha',
            'energy_electron',
            'energy_photon',
            'number_1',
            'number_2',
            'number_3',
            'number_4',
            'number_5',
            'atomic_mass',
            'air_kerma_rate_constant',
            'air_kerma_rate_coefficient'
        ]

        self.COLUMN_COL = [
            1,
            8,
            18,
            27,
            34,
            41,
            48,
            54,
            63,
            68,
            79,
            87,
            93,
            104,
            112,
            118,
            129,
            138,
            143,
            154,
            161,
            169,
            177,
            181,
            185,
            189,
            194,
            198,
            209,
            218
        ]

        self.nuclide_name = None
        self.halflife = None
        self.decay_mode = None
        self.pointer_1 = None
        self.pointer_2 = None
        self.pointer_3 = None
        self.pointer_4 = None
        self.daughter_nuclide_1_name = None
        self.daughter_nuclide_1_record_i = None
        self.daughter_nuclide_1_branch_i = None
        self.daughter_nuclide_2_name = None
        self.daughter_nuclide_2_record_i = None
        self.daughter_nuclide_2_branch_i = None
        self.daughter_nuclide_3_name = None
        self.daughter_nuclide_3_record_i = None
        self.daughter_nuclide_3_branch_i = None
        self.daughter_nuclide_4_name = None
        self.daughter_nuclide_4_record_i = None
        self.daughter_nuclide_4_branch_i = None
        self.energy_alpha = None
        self.energy_electron = None
        self.energy_photon = None
        self.number_1 = None
        self.number_2 = None
        self.number_3 = None
        self.number_4 = None
        self.number_5 = None
        self.atomic_mass = None
        self.air_kerma_rate_constant = None
        self.air_kerma_rate_coefficient = None

    def __repr__(self):
        return str(
            {
                column_name: getattr(self, column_name) 
                for column_name in self.COLUMN_NAMES
            }
        )

    def __extract_halflife(self):
        halflife_regex = r"([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)([a-z]+)"
        halflife_groups = re.match(halflife_regex, self.halflife, re.I).groups()
        return {
            'value': float(halflife_groups[0]),
            'unit': halflife_groups[2]
        }

    def get_nuclide(self):
        extracted_halflife = self.__extract_halflife()

        nuclide = Nuclide(
            name=self.nuclide_name,
            halflife_value=extracted_halflife.get('value'),
            halflife_unit=extracted_halflife.get('unit'),
            decay_mode=self.decay_mode,
            daughter_nuclide_names=[],
            daughter_nuclide_ratios=[]
        )

        if self.daughter_nuclide_1_name:
            nuclide.daughter_nuclide_names.append(self.daughter_nuclide_1_name)
            nuclide.daughter_nuclide_ratios.append(float(self.daughter_nuclide_1_branch_i))
        if self.daughter_nuclide_2_name:
            nuclide.daughter_nuclide_names.append(self.daughter_nuclide_2_name)
            nuclide.daughter_nuclide_ratios.append(float(self.daughter_nuclide_2_branch_i))
        if self.daughter_nuclide_3_name:
            nuclide.daughter_nuclide_names.append(self.daughter_nuclide_3_name)
            nuclide.daughter_nuclide_ratios.append(float(self.daughter_nuclide_3_branch_i))
        if self.daughter_nuclide_4_name:
            nuclide.daughter_nuclide_names.append(self.daughter_nuclide_4_name)
            nuclide.daughter_nuclide_ratios.append(float(self.daughter_nuclide_4_branch_i))

        return nuclide


class DataLoader:
    def __init__(self, filename):
        self.filename = filename
        self.nuclides = {}

    def get_nuclides_dict(self):
        with open(self.filename, encoding='latin-1') as ro:
            for i, line in enumerate(ro.readlines()):
                # skip first line (no data)
                if i > 0:
                    self.__process_line(line)
        return self.nuclides
        
    def __process_line(self, line):
        row = DataRow()

        for i, item_start_col in enumerate(row.COLUMN_COL):
            item_end_col = row.COLUMN_COL[i+1] if (i + 1) < len(row.COLUMN_COL) else 10000
            column_name = row.COLUMN_NAMES[i]
            line_item = line[item_start_col - 1 : item_end_col - 1]
            setattr(row, column_name, line_item.strip())
        
        nuclide_name = getattr(row, row.COLUMN_NAMES[0])
        nuclide = row.get_nuclide()

        self.nuclides[nuclide_name] = nuclide



from decaychain.BatemanDecay import bateman_trial
from decaychain.FormatInput import format_input


def bateman_multi(dict_input, user_time, user_tunit):
    results = {}
    formatted = format_input(dict_input)

    for nuclide, activity in formatted.items():
        user_nuclide = str(nuclide)
        user_activity = float(activity)
        result = bateman_trial(user_nuclide, user_time, user_tunit, user_activity)
        for nuclide, value2 in result.items():
            if not results.get(nuclide):
                results[nuclide] = 0
            results[nuclide] += float(value2)
    return results


## Brings in the Unit Registry from the pint module.
## This adds the definitions for years as 'y' as well as 'yr' from mydef.txt which is stored in the proj folder
from pint import UnitRegistry
ureg = UnitRegistry()
ureg.define('ky = yr * 1000 = [time]')
ureg.define('My = yr * 1000000 = [time]')
ureg.define('Gy = yr * 1000000000 = [time]')
ureg.define('Ty = yr * 1000000000000 = [time]')
ureg.define('Py = yr * 1000000000000000 = [time]')
ureg.define('y  = yr = [time]')
ureg.define('m  = second * 60 = [time]')
ureg.define('h  = second * 3600 = [time]')

class Base:
    def __init__(self, value: float, unit: str):
        self.value = value
        self.unit = unit
        self.quantity = ureg.Quantity(value, unit)

    def __json__(self, request):
        return {
            'value': self.value,
            'unit': self.unit
        }

class Concentration(Base):
    def __init__(self, value, unit):
        super().__init__(value, unit)

class Halflife(Base):
    def __init__(self, value, unit):
        super().__init__(value, unit)

class Time(Base):
    def __init__(self, value, unit):
        super().__init__(value, unit)



from typing import List

from decaychain.MeasurementUnit import Halflife


class Nuclide:
    def __init__(
        self, 
        name: str = None, 
        radioactive: bool = True,
        halflife_value: float = None,
        halflife_unit: str = None,
        decay_mode: str = None,
        daughter_nuclide_names: List[str] = [],
        daughter_nuclide_ratios:  List[str] = []
    ):
        self.name = name
        self.radioactive = radioactive
        if halflife_value and halflife_unit:
            self.halflife = Halflife(halflife_value, halflife_unit)
        else:
            self.halflife = None
        self.decay_mode = decay_mode
        self.daughter_nuclide_names = daughter_nuclide_names
        self.daughter_nuclide_ratios = daughter_nuclide_ratios

    def __json__(self, request):
        return {
            'name': self.name,
            'radioactive': self.radioactive,
            'halflife': self.halflife,
            'decay_mode': self.decay_mode,
            'daughter_nuclide_names': self.daughter_nuclide_names,
            'daughter_nuclide_ratios': self.daughter_nuclide_ratios
        }



## For thThis function removes extra carriage returns, formats the nuclides
# and splits the input whitespace (dict divide) and \r\n (next item)

from decaychain.NuclideFormatter import NuclideFormatter

def format_input(inputs):
    outputs = {}
    ## Splits digits and text for nuclides and reassembles in format XY-123
    for nuclide, activity in inputs.items():
        nuclide_out = NuclideFormatter(nuclide)
        formatted = {str(nuclide_out): float(activity)}
        ## Adds the formatted nuclide to the outputs dictionary. Duplicates are summed.
        for k in formatted.keys():
            outputs[k] = outputs.get(k,0) + formatted[k]

    return outputs




import re

def NuclideFormatter(nuclide):
    ## Formats a nuclide into the correct format for ICRP-07 matching.
    user_nuclide = nuclide
    ## finds the atomic mass
    atomic_mass = int("".join(filter(str.isdigit, user_nuclide)))
    ## regular expressions to find m and find element which must be in format "Aa".
    meta_regex = r"[0-9]+m"
    element_regex = r"[A-Z][a-z]*"
    element = re.findall(element_regex, user_nuclide)[0]
    meta = re.search(meta_regex, user_nuclide)
    element_f = element[0].upper() + element[1:].lower()
    nuclide_out = str(element_f) + "-" + str(atomic_mass) + ('m' if meta else '')
    return nuclide_out


from decaychain.NuclideFormatter import NuclideFormatter
from decaychain.DecayChain import Generator


def HaflLifeSearch(Nuclide):
## Grabs the user input and renames
    user_nuclide = NuclideFormatter(Nuclide)

    # Finds the halflife using the Chain Generator Function
    chain_generator = Generator()
    nuclidess = chain_generator.nuclides_dict.get(user_nuclide)
    hl = nuclidess.halflife.value
    hl_unit = nuclidess.halflife.unit

    ## formats for return to fucntion that called it.
    half_life_result = [hl, hl_unit]
    return half_life_result


import json
from decaychain.MeasurementUnit import Concentration
from decaychain.MeasurementUnit import Time
from decaychain.Nuclide import Nuclide
from decaychain.DecayChain import Generator


class Request:
    def __init__(self, request_dict):
        self.measurements = request_dict.get('measurements')
        self.target_time_value = request_dict.get('target_time_value')
        self.target_time_unit = request_dict.get('target_time_unit')

class Result:
    def __init__(self, nuclide: Nuclide, concentration: Concentration, time: Time):
        self.nuclide = nuclide
        self.concentration = concentration
        self.time = time

    def __json__(self, request):
        return {
            'nuclide': self.nuclide,
            'calculation': {
                'concentration': self.concentration,
                'time': self.time
            }
        }

class Calculator:
    def __init__(self):
        self.chain_generator = Generator()

    def calculate(self, decay_calculation_request: Request) -> [Result]:
        result = []

        target_time = Time(
            value=decay_calculation_request.target_time_value,
            unit=decay_calculation_request.target_time_unit
        )

        for measurement in decay_calculation_request.measurements:
            nuclide_name = measurement.get('nuclide_name')
            chains = self.chain_generator.get_for_nuclide_name(nuclide_name)
            chains_names_only = [
                chain.get_nuclide_names()
                for chain in chains
            ]

            print(json.dumps(chains_names_only, indent=4))

            initial_concentration = Concentration(
                value=measurement.get('concentration_value'),
                unit=measurement.get('concentration_unit')
            )

            calculated_concentration = nuclide.calculate_concentration_at_time(initial_concentration, target_time)

            result.append(
                Result(
                    nuclide=nuclide,
                    concentration=calculated_concentration,
                    time=target_time
                )
            )

        return result



from decaychain.DecayChain import Generator
from decaychain.MeasurementUnit import Concentration
from decaychain.MeasurementUnit import Time

## Bateman equation calculator from module
from batemaneq import bateman_parent
from math import log as ln # required for bateman module to function

# Unit handling module
from pint import UnitRegistry
ureg = UnitRegistry()

def bateman_trial(nuclide, time, tunit, conc):

    chain_generator = Generator()

    nuclide_name = nuclide
    concentration_value = conc
    decay_time = Time(time, tunit).quantity
    decay_time.ito(ureg.years)

    # generate the decay chains for the supplied nuclide name
    chains = chain_generator.get_for_nuclide_name(nuclide_name)

    ## opens a final results list
    final_result = {}

    output_result = {}

    ## loop over decay chains and calculate
    for idx, chain in enumerate(chains):

        Thalf = []

        for item in chain.items:
            nuclide = item.nuclide
            ratio = item.ratio
            halflife = item.nuclide.halflife


            ## Builds the halflife chain for input into bateman, drops stable isotopes.
            if item.nuclide.halflife != None:
                x = item.nuclide.halflife.quantity
                x.ito(ureg.years)
                Thalf.append(x.magnitude)

        ## Runs results through bateman module.
        output_items = bateman_parent([ln(2) / x for x in Thalf], decay_time.magnitude)

        ## Converts the results output to final activity  and appends them to final_result dicionary

        for idx, output_item in enumerate(output_items):
            nuclide = chain.items[idx].nuclide
            if not final_result.get(nuclide.name):
                final_result[nuclide.name] = 0
            final_conc = output_item * (Thalf[0] / Thalf[idx]) * concentration_value
            relative_conc = final_conc * chain.ratio

            final_result[nuclide.name] += float(relative_conc)


    return final_result




from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements = [
"batemaneq>=0.2.2",
"Pint>=0.12",
"importlib_resources; python_version<'3.7'",
]

setup(
    name ='decaychain',
    version ='0.6.8',
    description='Module to radioactively decay radioactive elements using the ICRP-107 and Bateman Equation',
#   package_dir={''},
    license='MIT',
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Topic :: Scientific/Engineering",
        "Topic :: Education ",
        ],
    packages=find_packages(),
    author="Kenneth McKee",
    author_email="kenneth.mckee@protonmail.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Rolleroo/decaychain",
    python_requires='>=3.6',
    install_requires=requirements,
include_package_data = True,
package_data = {
'' : ['*.NDX'],
}
)


import copy
from decaychain.DataLoader import DataLoader
from decaychain.Nuclide import Nuclide
from decaychain.MeasurementUnit import Concentration
try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources


class Chain():
    def __init__(self):
        self.items = []
        self.ratio = 1

    def get_nuclide_names(self):
        return [
            item.nuclide.name
            for item in self.items
        ]

class Item():
    def __init__(self, nuclide: Nuclide, ratio: float, concentration: Concentration = None):
        self.nuclide = nuclide
        self.ratio = ratio
        self.concentration = concentration

class Generator():
    def __init__(self):
        path = 'data'
        with pkg_resources.path(__package__+'.'+path, 'ICRP-07.NDX') as path2:
            path2 = path2
        loader = DataLoader(path2)
        self.nuclides_dict = loader.get_nuclides_dict()
    
    def get_for_nuclide_name(self, nuclide_name):
        chains = []

        def __process_nuclide(nuclide, chain=Chain(), ratio=1):
            item = Item(
                nuclide=nuclide,
                ratio=ratio
            )
            chain.items.append(item)

            chain.ratio = ratio * chain.ratio

            for idx, daughter_nuclide_name in enumerate(nuclide.daughter_nuclide_names):
                daughter_nuclide_ratio = nuclide.daughter_nuclide_ratios[idx]
                daughter_nuclide = self.nuclides_dict.get(daughter_nuclide_name)

                if daughter_nuclide:
                    # the daughter nuclide was found in the decay list, so continue
                    new_chain = copy.deepcopy(chain)
                    __process_nuclide(daughter_nuclide, new_chain, ratio=daughter_nuclide_ratio)
                elif idx == 0:
                    # we reached the end of the chain
                    final_chain = copy.deepcopy(chain)

                    stable_nuclide = Nuclide(
                        name=daughter_nuclide_name,
                        radioactive=False
                    )

                    final_chain.items.append(
                        Item(
                            nuclide=stable_nuclide,
                            ratio=None
                        )      
                    )

                    chains.append(final_chain)

        nuclide = self.nuclides_dict.get(nuclide_name)
        if not nuclide:
            return chains

        __process_nuclide(nuclide)

        return chains